@@ -1,7 +1,7 @@
 package App::fatten;
 
-our $DATE = '2015-04-11'; # DATE
-our $VERSION = '0.33'; # VERSION
+our $DATE = '2015-04-15'; # DATE
+our $VERSION = '0.34'; # VERSION
 
 use 5.010001;
 use strict;
@@ -11,17 +11,48 @@ use Log::Any '$log';
 use Log::Any::For::Builtins qw(system my_qx);
 BEGIN { no warnings; $main::Log_Level = 'info' }
 
-use App::tracepm ();
-
+use App::tracepm (); # we need list of trace methods too so we load early
 use File::chdir;
 use File::Slurper qw(write_text read_text);
 use version;
 
-sub _sq {
+sub __sq {
     require String::ShellQuote;
     String::ShellQuote::shell_quote($_[0]);
 }
 
+sub _run_lcpan {
+    require IPC::System::Options;
+    require JSON;
+
+    my $self = shift;
+
+    state $checked;
+    unless ($checked) {
+        require File::Which;
+        File::Which::which("lcpan")
+            or die "lcpan is not available, please install it first\n";
+        my $res = IPC::System::Options::backtick(
+            {die=>1, log=>1}, "lcpan", "stats", "--json", "--no-naked-res");
+        $res = JSON::decode_json($res);
+        die "Can't 'lcpan stat': $res->[0] - $res->[1]\n"
+            unless $res->[0] == 200;
+        my $stats = $res->[2];
+        if ((time - $stats->{raw_last_index_time}) > 5*86400) {
+            die "lcpan index is over 5 days old, please refresh it first ".
+                "with 'lcpan update'\n";
+        }
+    }
+
+    my @lcpan_args = (@_, "--json", "--no-naked-res");
+    my $res = IPC::System::Options::backtick(
+        {die=>1, log=>1}, "lcpan", @lcpan_args);
+    $res = JSON::decode_json($res);
+    die "Can't 'lcpan ".join(" ", @lcpan_args)."': $res->[0] - $res->[1]\n"
+        unless $res->[0] == 200;
+    $res->[2];
+}
+
 our %SPEC;
 
 sub _trace {
@@ -30,19 +61,22 @@ sub _trace {
     return if $self->{trace_method} eq 'none';
 
     $log->debugf("  Tracing with method '%s' ...", $self->{trace_method});
-    my $res = App::tracepm::tracepm(
+    my %traceargs = (
         method => $self->{trace_method},
         script => $self->{input_file},
         args => $self->{args},
         use => $self->{use},
         recurse_exclude_core => $self->{exclude_core} ? 1:0,
-        recurse_exclude_xs   => 1,
         detail => 1,
+        trap_script_output => 1,
 
         core => $self->{exclude_core} ? 0 : undef,
-        xs   => 0,
+        ($self->{trace_extra_opts} ? %{$self->{trace_extra_opts}} : ()),
     );
-    die "Can't trace: $res->[0] - $res->[1]" unless $res->[0] == 200;
+
+    $log->debugf("  tracepm args: %s", \%traceargs);
+    my $res = App::tracepm::tracepm(%traceargs);
+    die "Can't trace: $res->[0] - $res->[1]\n" unless $res->[0] == 200;
     $self->{deps} = $res->[2];
 }
 
@@ -66,7 +100,6 @@ sub _build_lib {
     my $deps = $self->{deps};
     for (@$deps) {
         next if $_->{is_core} && $self->{exclude_core};
-        next if $_->{is_xs};
         $log->debugf("  Adding module: %s (traced)", $_->{module});
         $mod_paths{$_->{module}} = undef;
     }
@@ -87,10 +120,10 @@ sub _build_lib {
         }
     }
 
-    if (defined(my $file = $self->{include_from_list})) {
+    if (defined(my $file = $self->{include_list})) {
         $log->debugf("  Adding modules listed in: %s", $file);
         open my($fh), "<", $file
-            or die "Can't open modules list file '$file': $!";
+            or die "Can't open modules list file '$file': $!\n";
         my $linenum = 0;
         while (my $line = <$fh>) {
             $linenum++;
@@ -110,7 +143,7 @@ sub _build_lib {
         }
     }
 
-    for (@{ $self->{include_from_dir} // [] }) {
+    for (@{ $self->{include_dir} // [] }) {
         $log->debugf("  Adding modules found in: %s", $_);
         local $CWD = $_;
         File::Find::find(
@@ -132,6 +165,11 @@ sub _build_lib {
     my %fmod_paths; # filtered mods
   MOD:
     for my $mod (sort keys %mod_paths) {
+        if ($self->{_exclude_deps} && $self->{_exclude_deps}{$mod}) {
+            $log->infof("Excluding %s: skipped by allow_dep %s", $mod, $self->{_exclude_deps}{$mod});
+            next MOD;
+        }
+
         if ($self->{exclude} && $mod ~~ @{ $self->{exclude} }) {
             $log->infof("Excluding %s: skipped", $mod);
             next MOD;
@@ -154,12 +192,12 @@ sub _build_lib {
                 next MOD;
             }
         }
-        if (defined(my $file = $self->{exclude_from_list})) {
+        if (defined(my $file = $self->{exclude_list})) {
             if (!$excluded_list) {
                 $excluded_list = [];
                 $log->debugf("  Reading excludes listed in: %s", $file);
                 open my($fh), "<", $file
-                    or die "Can't open modules list file '$file': $!";
+                    or die "Can't open modules list file '$file': $!\n";
                 my $linenum = 0;
                 while (my $line = <$fh>) {
                     $linenum++;
@@ -183,17 +221,25 @@ sub _build_lib {
     }
     %mod_paths = %fmod_paths;
 
+    require Module::XSOrPP;
     for my $mod (sort keys %mod_paths) {
-        my $mpath = $mod_paths{$mod} //
-            Module::Path::More::module_path(module=>$mod);
-        defined $mpath or die "Can't find path for $mod";
+        my $mpath = $mod_paths{$mod};
+
+        unless ($mpath) {
+            if (Module::XSOrPP::is_xs($mod)) {
+                die "Can't add XS module: $mod\n";
+            }
+        }
+
+        $mpath //= Module::Path::More::module_path(module=>$mod);
+        defined $mpath or die "Can't find path for $mod\n";
 
         my $modp = $mod; $modp =~ s!::!/!g; $modp .= ".pm";
         my ($dir) = $modp =~ m!(.+)/(.+)!;
         if ($dir) {
             my $dir_to_make = "$tempdir/lib/$dir";
             unless (-d $dir_to_make) {
-                File::Path::make_path($dir_to_make) or die "Can't make_path: $dir_to_make";
+                File::Path::make_path($dir_to_make) or die "Can't make_path: $dir_to_make\n";
             }
         }
 
@@ -246,10 +292,10 @@ sub _pack {
     system join(
         "",
         "fatpack file ",
-        _sq($self->{abs_input_file}), " > ",
-        _sq($self->{abs_output_file}),
+        __sq($self->{abs_input_file}), " > ",
+        __sq($self->{abs_output_file}),
     );
-    die "Can't fatpack file: ".Proc::ChildError::explain_child_error() if $?;
+    die "Can't fatpack file: ".Proc::ChildError::explain_child_error()."\n" if $?;
 
     chmod 0755, $self->{abs_output_file};
 
@@ -268,6 +314,43 @@ sub _pack {
                 $self->{abs_output_file}, (-s $self->{abs_output_file})/1024);
 }
 
+sub _test {
+    require Capture::Tiny;
+    require IPC::System::Options;
+
+    my $self = shift;
+    die "Can't test: at least one test case ('--test-case-json') must be specified\n"
+        unless $self->{test_cases} && @{ $self->{test_cases} };
+
+    my $cases = $self->{test_cases};
+    my $i = 0;
+    for my $case (@$cases) {
+        $i++;
+        $log->debugf("  Test case %d/%d: %s ...", $i, ~~@$cases, $case->{args});
+        my @cmd = ($^X);
+        push @cmd, @{ $case->{perl_args} } if $case->{perl_args} && @{ $case->{perl_args} };
+        push @cmd, $self->{abs_output_file}, @{ $case->{args} };
+        my $exit;
+        # log statement by IPC::System::Options' log=1 will be eaten by
+        # Capture::Tiny, so we log here
+        $log->tracef("cmd: %s", \@cmd);
+        my $output = Capture::Tiny::capture_merged(
+            sub {
+                IPC::System::Options::system({log=>0, shell=>0}, @cmd);
+                $exit = $?;
+            }
+        );
+        my $expected_exit = $case->{exit_code} // 0;
+        if ($exit != $expected_exit) {
+            die "  Test case $i failed: exit code is not $expected_exit ($exit),output: <<$output>>\n";
+        }
+        if (defined $case->{output_like}) {
+            $output =~ /$case->{output_like}/
+                or die "  Test case $i failed: output does not match $case->{output_like}, output: <<$output>>\n";
+        }
+    }
+}
+
 sub new {
     my $class = shift;
     bless { @_ }, $class;
@@ -334,13 +417,13 @@ _
             tags => ['category:module-selection'],
             'x.schema.element_entity' => 'modulename',
         },
-        include_from_list => {
+        include_list => {
             summary => 'Include extra modules from a list in a file',
             schema => 'str*', # XXX filename
             tags => ['category:module-selection'],
             'x.schema.entity' => 'filename',
         },
-        include_from_dir => {
+        include_dir => {
             summary => 'Include extra modules under directories',
             'summary.alt.plurality.singular' => 'Include extra modules under a directory',
             schema => ['array*' => of => 'str*'],
@@ -407,7 +490,7 @@ _
             schema => ['bool' => default => 1],
             tags => ['category:module-selection'],
         },
-        exclude_from_list => {
+        exclude_list => {
             summary => 'Exclude modules from a list in a file',
             schema => 'str*', # XXX filename
             tags => ['category:module-selection'],
@@ -453,6 +536,31 @@ _
             cmdline_aliases => { t=>{} },
             tags => ['category:module-selection'],
         },
+        trace_extra_opts => {
+            schema => ['hash*'],
+            summary => 'Pass more options to `App::tracepm`',
+            tags => ['category:module-selection'],
+        },
+        allow_deps => {
+            'summary.alt.plurality.singular' => 'Allow script to depend on a module instead of fatpacking it',
+            schema => ['array*', of=>'str*'],
+            'x.name.is_plural' => 1,
+            description => <<'_',
+
+This option can be used to express that script will depend on a specified
+module. The distribution of that module and the recursive dependencies for _that
+distribution_ will be retrieved recursively using `lcpan`. All these will be
+excluded from being fatpacked.
+
+This option can be used to express dependency to an XS module, since a script
+cannot fatpack XS modules.
+
+This option requires that `lcpan` is installed and a fairly recent lcpan index
+is available.
+
+_
+            tags => ['category:module-selection'],
+        },
         use => {
             summary => 'Additional modules to "use"',
             'summary.alt.plurality.singular' => 'Additional module to "use"',
@@ -468,6 +576,7 @@ _
         },
         args => {
             summary => 'Script arguments',
+            'x.name.is_plural' => 1,
             'summary.alt.plurality.singular' => 'Script argument',
             description => <<'_',
 
@@ -577,6 +686,33 @@ _
             schema => ['bool' => default=>0],
             tags => ['category:debugging'],
         },
+
+        test => {
+            schema => ['bool', is=>1],
+            summary => 'Test the resulting output',
+            cmdline_aliases => {T=>{}},
+            description => <<'_',
+
+Testing is done by running the resulting fatpacked result with `perl
+-Mlib::core::only`. To test, at least one test case is required (see
+`--test-case-json`). Test cases specify what arguments to give to program, what
+exit code we expect, and what the output should contain.
+
+_
+            tags => ['category:testing'],
+        },
+        test_cases => {
+            schema => ['array*', of=>'hash*'],
+            'x.name.is_plural' => 1,
+            description => <<'_',
+
+Example case:
+
+    {"args":["--help"], "exit_code":0, "perl_args":["-Mlib::core::only"], "output_like":"Usage:"}
+
+_
+            tags => ['category:testing'],
+        },
     },
     deps => {
         exec => 'fatpack',
@@ -591,6 +727,24 @@ sub fatten {
     my %args = @_;
     my $self = __PACKAGE__->new(%args);
 
+    if ($self->{allow_deps} && @{ $self->{allow_deps} }) {
+        my %exclude_deps;
+        for my $dep (@{ $self->{allow_deps} }) {
+            my @dep_mods = ($dep);
+            my $res = $self->_run_lcpan("deps", "-R", $dep);
+            for my $ent (@{ $res }) {
+                $ent->{module} =~ s/^\s+//;
+                push @dep_mods, $ent->{module};
+            }
+            $res = $self->_run_lcpan("mods-from-same-dist", @dep_mods);
+            for my $ent (@{ $res }) {
+                $log->debugf("  Adding to exclude_deps: %s (allow_dep %s)", $ent, $dep);
+                $exclude_deps{$ent} = $dep;
+            }
+        }
+        $self->{_exclude_deps} = \%exclude_deps;
+    }
+
     my $tempdir = File::Temp::tempdir(CLEANUP => 0);
     $log->debugf("Created tempdir %s", $tempdir);
     $self->{tempdir} = $tempdir;
@@ -695,6 +849,11 @@ sub fatten {
         unlink $self->{abs_output_file};
     }
 
+    if ($self->{test}) {
+        $log->infof("Testing ...");
+        $self->_test;
+    }
+
     [200];
 }
 
@@ -713,7 +872,7 @@ App::fatten - Pack your dependencies onto your script file
 
 =head1 VERSION
 
-This document describes version 0.33 of App::fatten (from Perl distribution App-fatten), released on 2015-04-11.
+This document describes version 0.34 of App::fatten (from Perl distribution App-fatten), released on 2015-04-15.
 
 =head1 SYNOPSIS
 
@@ -738,6 +897,19 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
+=item * B<allow_deps> => I<array[str]>
+
+This option can be used to express that script will depend on a specified
+module. The distribution of that module and the recursive dependencies for I<that
+distribution> will be retrieved recursively using C<lcpan>. All these will be
+excluded from being fatpacked.
+
+This option can be used to express dependency to an XS module, since a script
+cannot fatpack XS modules.
+
+This option requires that C<lcpan> is installed and a fairly recent lcpan index
+is available.
+
 =item * B<args> => I<array[str]>
 
 Script arguments.
@@ -774,7 +946,7 @@ Just like the C<exclude> option, but will exclude module as well as other module
 from the same distribution. Module name must be the main module of the
 distribution. Will determine other modules from the C<.packlist> file.
 
-=item * B<exclude_from_list> => I<str>
+=item * B<exclude_list> => I<str>
 
 Exclude modules from a list in a file.
 
@@ -791,6 +963,10 @@ Include extra modules.
 When the tracing process fails to include a required module, you can add it
 here.
 
+=item * B<include_dir> => I<array[str]>
+
+Include extra modules under directories.
+
 =item * B<include_dist> => I<array[str]>
 
 Include all modules of dist.
@@ -799,11 +975,7 @@ Just like the C<include> option, but will include module as well as other module
 from the same distribution. Module name must be the main module of the
 distribution. Will determine other modules from the C<.packlist> file.
 
-=item * B<include_from_dir> => I<array[str]>
-
-Include extra modules under directories.
-
-=item * B<include_from_list> => I<str>
+=item * B<include_list> => I<str>
 
 Include extra modules from a list in a file.
 
@@ -883,6 +1055,25 @@ Set strip_ws=1 (strip whitespace) in Perl::Stripper.
 
 Only relevant when stripping using Perl::Stripper.
 
+=item * B<test> => I<bool>
+
+Test the resulting output.
+
+Testing is done by running the resulting fatpacked result with C<perl
+-Mlib::core::only>. To test, at least one test case is required (see
+C<--test-case-json>). Test cases specify what arguments to give to program, what
+exit code we expect, and what the output should contain.
+
+=item * B<test_cases> => I<array[hash]>
+
+Example case:
+
+ {"args":["--help"], "exit_code":0, "perl_args":["-Mlib::core::only"], "output_like":"Usage:"}
+
+=item * B<trace_extra_opts> => I<hash>
+
+Pass more options to `App::tracepm`.
+
 =item * B<trace_method> => I<str> (default: "fatpacker")
 
 Which method to use to trace dependencies.
